/*                                            
Configuration.cc
*/

#include "Configuration.h"

Configuration::Configuration (const char *afnconf, bool aAbortInError)
  {
  fnconf = afnconf;
  AbortInError = aAbortInError;
  }
  
Configuration::~Configuration ()
  {
  }

void Configuration::Error (void)
  {
  if (AbortInError == true) 
    {
    printf ("Error reading configuration file \"%s\"! Aborting ...\n", fnconf.cstr()); 
    exit (1);
    }
  }  

void Configuration::setFileName (const char *afilename)
  {
  fnconf = afilename;
  }
      
const char *Configuration::getFileName (void)
  {
  return fnconf.cstr();
  }

bool Configuration::reload (void)
  {
  return ConfRead ();
  }    

//METO TODO EL FICHERO DE CONFIGURACION EN MEMORIA
bool Configuration::ConfRead (void)
  {
  char Buffer[XXBUFFERSIZE + 1];
  FILE *CfgFile;
  SLFile.Clear();
  CfgFile = fopen (fnconf.cstr(), "r");
  if (NULL == CfgFile) {Error(); return false;}
  while (NULL != fgets (Buffer, XXBUFFERSIZE, CfgFile))
    {
    if (IsEmpty(Buffer) == false) SLFile.Add (Buffer);              
    }
  fclose (CfgFile);
  return true;
  }
  
//GUARDO EN DISCO TODO EL FICHERO DE CONFIGURACION DE MEMORIA
bool Configuration::ConfSave (void)
  {
  TSBuffer abuf;
  FILE *CfgFile;
  CfgFile = fopen (fnconf.cstr(), "w");
  if (NULL == CfgFile) {Error(); return false;}
  for (int yy = 0; yy < SLFile.Count(); ++yy)
    {
    strncpy (abuf, SLFile.getString(yy).cstr(), CSMALLBUFFER);
    QuitaRetornoCarroDeLinea (abuf);
    if (IsEmpty(abuf) == false) fprintf (CfgFile, "%s\n", abuf);
    }
  fclose (CfgFile);
  return true;
  }  

bool Configuration::newFile (const char *filecont)
  {
  FILE *CfgFile;
  CfgFile = fopen (fnconf.cstr(), "w");
  if (NULL == CfgFile) {Error(); return false;}
  fprintf (CfgFile, "%s", filecont);
  fclose (CfgFile);
  return true;  
  }    
 
void Configuration::dumpDebug (void)
  {
  for (int yy = 0; yy < SLFile.Count(); ++yy)
    {
    DEBUG ("%s", SLFile.getString(yy).cstr());  //of DEBUG, MUST BE UNCOMMENTED
    }
  }
 
//RETURNS A CONFIGURATION STRING. LIKE AS getprivateprofilestring OF WINDOWS API
char *Configuration::getString (const char *asection, const char *akey, const char *adefault, int amaxsize)
  {
  bool EncontradaSeccion = false;
  int i=0, maxsize;
  char cbegin, cend, *pos;
  TBuffer linea, seccion, key, result, upseccion, upkey;
  
  maxsize = amaxsize; if (maxsize > CMAXBUFFER) {maxsize = CMAXBUFFER;}
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  strncpy (upkey, akey, CMAXBUFFER); xucase (upkey);
  initStr(seccion); initStr (key); initStr (result);
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {EncontradaSeccion = true;}
          else {EncontradaSeccion = false;}
          }
        else
          {
          if (EncontradaSeccion == true)
            {
            pos = strchr (linea, '=');
            if (pos != NULL)
              {
              strncpy (key, linea, CMAXBUFFER); 
              key [pos - linea] = '\0'; trim (key); xucase (key);
              if (strcmp (key, upkey) == 0)
                {
                strncpy (result, pos + 1, maxsize); trim (result);
                }
              }
            }
          }  
        }
      }
    ++i;
    }
  if (strcmp (result, "") == 0) {strncpy (result, adefault, CMAXBUFFER);}  
  /*
  DEBUG ("*******************************");
  DEBUG ("SECCION=%s", asection);
  DEBUG ("KEY=%s", akey);
  DEBUG ("DEFAULT=%s", adefault);
  DEBUG ("RESULT=%s", result);
  */
  return xstrdup (result);  
  }  
  
bool Configuration::setString (const char *asection, const char *akey, const char *astring, int amaxsize)
  {
  bool EncontradaSeccion = false;
  int i=0, maxsize;
  char cbegin, cend, *pos;
  TBuffer linea, seccion, key, result, anewline, upseccion, upkey;
  
  maxsize = amaxsize; if (maxsize > CMAXBUFFER) {maxsize = CMAXBUFFER;}
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  strncpy (upkey, akey, CMAXBUFFER); xucase (upkey);
  initStr(seccion); initStr (key); initStr (result);
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {
          if (EncontradaSeccion == true)
            {
            //add a new line for new variable not in the ini file
            sprintf (anewline, "%s=%s", akey, astring);
            SLFile.insertElementAt (anewline, i);
            return true;
            }          
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {EncontradaSeccion = true;}
          else {EncontradaSeccion = false;}
          }
        else
          {
          if (EncontradaSeccion == true)
            {
            pos = strchr (linea, '=');
            if (pos != NULL)
              {
              strncpy (key, linea, CMAXBUFFER); 
              key [pos - linea] = '\0'; trim (key); xucase (key);
              if (strcmp (key, upkey) == 0)
                {
                strncpy (result, pos + 1, maxsize); trim (result);
                sprintf (linea, "%s=%s", akey, astring);
                SLFile.setElementAt (linea, i); 
                return true;
                }
              }
            }
          }  
        }
      }
    ++i;
    }
 if (EncontradaSeccion == true)
   {
   //add a new line for new variable not in the ini file
   sprintf (anewline, "%s=%s", akey, astring);
   SLFile.insertElementAt (anewline, i - 1);
   return true;
   }     
   
  if (IsThereSection (asection) == false)
    {
    sprintf (linea, "[%s]", asection);       SLFile.Add (linea);
    sprintf (linea, "%s=%s", akey, astring); SLFile.Add (linea);
    }
     
  return false;  
  }  
  
//DEVUELVE UN INTEGER DE CONFIGURATION. 
int Configuration::getInt (const char *asection, const char *akey, int adefault)
  {
  bool EncontradaSeccion = false;
  int i=0, maxsize, resint;
  char cbegin, cend, *pos;
  TBuffer linea, seccion, key, result, upseccion, upkey;
  
  maxsize = CMAXBUFFER;
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  strncpy (upkey, akey, CMAXBUFFER); xucase (upkey);
  initStr(seccion); initStr (key); initStr (result);
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {EncontradaSeccion = true;}
          else {EncontradaSeccion = false;}
          }
        else
          {
          if (EncontradaSeccion == true)
            {
            pos = strchr (linea, '=');
            if (pos != NULL)
              {
              strncpy (key, linea, CMAXBUFFER); 
              key [pos - linea] = '\0'; trim (key); xucase (key);
              if (strcmp (key, upkey) == 0)
                {
                strncpy (result, pos + 1, maxsize); trim (result);
                }
              }
            }
          }  
        }
      }
    ++i;
    }
  resint = xatoidef (result, adefault);
  /*
  DEBUG ("*******************************");
  DEBUG ("SECCION=%s", asection);
  DEBUG ("KEY=%s", akey);
  DEBUG ("DEFAULT=%d", adefault);
  DEBUG ("RESULT=%d", resint);
  */
  return resint;  
  }   
  
bool Configuration::setInt (const char *asection, const char *akey, int aint)
  {
  bool EncontradaSeccion = false;
  int i=0, maxsize;
  char cbegin, cend, *pos;
  TBuffer linea, seccion, key, result, anewline, upseccion, upkey;
  
  maxsize = CMAXBUFFER;
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  strncpy (upkey, akey, CMAXBUFFER); xucase (upkey);
  initStr(seccion); initStr (key); initStr (result);
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {          
          if (EncontradaSeccion == true)
            {
            //add a new line for new variable not in the ini file
            sprintf (anewline, "%s=%d", akey, aint);
            SLFile.insertElementAt (anewline, i);
            return true;
            }          
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {EncontradaSeccion = true;}
          else {EncontradaSeccion = false;}
          }
        else
          {
          if (EncontradaSeccion == true)
            {
            pos = strchr (linea, '=');
            if (pos != NULL)
              {
              strncpy (key, linea, CMAXBUFFER); 
              key [pos - linea] = '\0'; trim (key); xucase (key);
              if (strcmp (key, upkey) == 0)
                {
                strncpy (result, pos + 1, maxsize); trim (result);
                sprintf (linea, "%s=%d", akey, aint);
                SLFile.setElementAt (linea, i); 
                return true;
                }
              }
            }
          }  
        }
      }
    ++i;
    }
  if (EncontradaSeccion == true)
    {
    //add a new line for new variable not in the ini file
    sprintf (anewline, "%s=%d", akey, aint);
    SLFile.insertElementAt (anewline, i - 1);
    return true;
    }            
    
  if (IsThereSection (asection) == false)
    {
    sprintf (linea, "[%s]", asection);       SLFile.Add (linea);
    sprintf (linea, "%s=%d", akey, aint); SLFile.Add (linea);
    }
     
  return false;  
  }     
  
//EXISTE SECCION
bool Configuration::IsThereSection (const char *asection)
  {
  int i=0;
  char cbegin, cend;
  TBuffer seccion, linea, upseccion;
  
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  initStr(seccion); 
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {return true;}
          }
        }
      }
    ++i;
    }
  return false;  
  }    
  
//COGE EL CONTENIDO DE UNA SECCION ENTERA
bool Configuration::getSection (const char *asection, StringList *ASection)
  {
  bool HabiaSeccion = false, EncontradaSeccion = false;
  int i=0;
  char cbegin, cend;
  TBuffer seccion, linea, upseccion;
  
  ASection->Clear();
  strncpy (upseccion, asection, CMAXBUFFER); xucase (upseccion);
  initStr(seccion); 
  while (i < SLFile.Count())
    {
    strncpy (linea, SLFile.getString(i).trim().cstr(), CMAXBUFFER);
    if ((EncontradaSeccion == true) && (linea != NULL))
      {
      if (linea[0] != '[') {ASection->Add (linea);}
      }
    if ((strcmp (linea, "") != 0) && (linea != NULL) && (strcmp (linea, " ") != 0))
      {
      cbegin = linea[0]; cend = linea[strlen(linea) - 1];
      if (cbegin != ';')
        {
        if ((cbegin == '[') && (cend == ']'))
          {
          XString axs (SLFile.getString(i).trim());
          strncpy (seccion, axs.substring(1, axs.length() - 1).toUpperCase().cstr(), CMAXBUFFER);
          if (strcmp (seccion, upseccion) == 0) {HabiaSeccion = true; EncontradaSeccion = true;}
          else {EncontradaSeccion = false;}
          }
        }
      }
    ++i;
    }
  return HabiaSeccion;  
  }      
  
