/*                         
InterDaemon.cc
*/

/* ====================================================================
 * Copyright (c) 2000 University of Valencia.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the University 
 *    of Valencia."
 *
 * 4. The names "Postman" must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the University of Valencia"
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include "InterDaemon.h"
#include "Config.h"
#include "Utils.h"
#include "XSocket.h"
#include "HTML.h"
#include "Language.h"
#include "CCLIENT.h"
#include "SendMSG.h"
#include "UserOptions.h"
#include "Configuration.h"
#include "Addressbook.h"
#include "rutinas_db.h"
#include "PsTitle.h"
#include "Cookie.h"
#include "ClientNNTP.h"

XSocket *serversocket;
HTML *html;
Language *L;
CCLIENT *cclient;
ClientNNTP *CN;
Configuration *Conf;
TBuffer UserDir, username, ipclient, lastconnid, stcmd, BasePath;
static int ServerCounter = 1;
static long int TotalLogins = 1;
pid_t MainServerPid;
int LockFileDescriptor, lastcmd;
Cookie *cookie;

int main  (int argc, char **argv, char **envp)
  {
  TBuffer linea, connid, stlang, pw, parm3, parm4, parm5, allparms, allparms2, imapserver, beginmailbox, tmpbuf,
          abuf, abuf2, amb;
  int cs, cmd, validator, pageid;
  char *fnsocket;
  UserOptions *uo;
  StringList *SLHeaders, *SLMailboxes;
  bool boo;
    
  initServer_beforeFork(argc, argv, envp, &cs);    
      
  //**************************************************
  //********************    FROM HERE THE NEW CHILD! *
  //**************************************************
  
  initServer_afterFork (argc, argv, envp, cs);
  sm_setproctitle (PROGNAME, true, "   (After fork)");

  html = new HTML (serversocket);   
  serversocket->Gets (linea, CMAXBUFFER);
  ini (connid, stlang, stcmd, allparms, allparms2, username, pw, parm3, parm4, parm5);
  initStr (imapserver);
  sscanf (linea, "%s %d %d %d %s %s %s %[^]]s", connid, &pageid, &validator, &lastcmd, stlang, ipclient, stcmd, allparms);
  xscanf (linea, connid, &pageid, &validator, &lastcmd, stlang, ipclient, stcmd, username, pw, imapserver, beginmailbox, NULL);
  if (IsEmpty(beginmailbox) == true) {strncpy (beginmailbox, DEFAULTBEGINMAILBOX, CMAXBUFFER);}
  
  L = new Language (stlang);
  cmd = TranslateCmd (stcmd);
  
  if (CONVERTUSERNAMETOLOWERCASE == 1) {xlcase (username);}  
  
  CheckServerCount (stlang);
  CheckParams (cmd, connid, pw);
  CheckAllowServeralConns ();      
  
  strncpy (lastconnid, connid, CMAXBUFFER);  
  strncpy (BasePath, Conf->getString ("main", "usersbasepath", "/var/postman/users", CMAXBUFFER), CMAXBUFFER);  
  uo = new UserOptions (BasePath, username, imapserver);
  uo->setPageId (pageid);
  uo->set_use_javascript (Conf->getInt ("main", "use_javascript", 1));
  html->setMAXMSGSFORPAGE (uo->getMsgsPerIndexPage());
  CN = new ClientNNTP (uo->getNNTPServer(), Conf->getString ("main", "defaultipdomain", "", CMAXIPNAME), 
                       username, pw);
  cclient = new CCLIENT ();
  cclient->setMAXMSGSFORPAGE (uo->getMsgsPerIndexPage());
  cclient->setTimeout (IMAP_TIMEOUT);
  sm_setproctitle (PROGNAME, true, "%s   (Before login)", username);  
  if (doLogin (uo, imapserver, pw) == true) 
    { 
    cclient->setNetNewsRC();
    sm_setproctitle (PROGNAME, true, "%s   (After login OK)", username);
    doLocking ();
    sm_setproctitle (PROGNAME, true, "%s   (After doLocking)", username);
    uo->InitStatusLine (); 
    if (cclient->getError()==true) {uo->setStatusLine (cclient->getErrorString(), RED);}
    setUserDirs(uo);    
    uo->leeConf();
    sm_setproctitle (PROGNAME, true, "%s   (After leeConf)", username);
    DelAllFilesFromDir ((char *)uo->getAttachDir());
    strncpy (connid, make_session_id (imapserver, tmpbuf), CMAXBUFFER);
    strncpy (lastconnid, connid, CMAXBUFFER);
    fnsocket = make_session_socket (username, connid, tmpbuf);
    SLHeaders = new StringList ();
    SLMailboxes = new StringList();
    cclient->expandMailboxName (beginmailbox, amb);
    boo = cclient->ChangeMailbox (amb);
    if (boo == true)
      {
      if (cclient->getServerProtocol() == PROTOCOL_NNTP)
        {
        strcpy (abuf, L->get(MSG_NEW_TO_FORUMS));
        uo->setStatusLine (abuf, RED);
        }        
      uo->setCurrentPosition (0, cclient->getTotalMsgs ());
      sm_setproctitle (PROGNAME, true, "%s   (After setCurrentPosition)", username);
      cclient->setMAXMSGSFORPAGE (uo->getMsgsPerIndexPage());
      sm_setproctitle (PROGNAME, true, "%s   (After setMAXMSGSFORPAGE)", username);
      cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO));
      sm_setproctitle (PROGNAME, true, "%s   (After getHeaderList)", username);
      cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, true);
      sm_setproctitle (PROGNAME, true, "%s   (After getMailboxes)", username);
      uo->setMailboxSize (cclient->getMailboxSize());
      uo->setPageId(0);
      xsnprintf (linea, CMAXBUFFER, "%s 0 0 0 %s %s mb_index\n", connid, stlang, ipclient);
      uo->addLastStringCmd (linea);
      uo->setMailboxSize (cclient->getMailboxSize());
      cookie = new Cookie (username);
      if (Conf->getInt ("main", "use_cookies", 0) == 1) {html->printCookie(cookie->getSetFullCookieLine());}
      html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
      }
    else 
      {
      if (cclient->getError() == true) 
        xsnprintf (abuf, CMAXBUFFER, cclient->getErrorString());
      else
        xsnprintf (abuf, CMAXBUFFER, L->get(ERR_INV_MAILBOX), cclient->getFullMailbox(amb, abuf2));
      html->ErrorPage(abuf);
      }
    delete SLHeaders;
    delete SLMailboxes;
    serversocket->Close();
    sm_setproctitle (PROGNAME, true, "%s   (Before ChildLoop)", username);
    ChildLoop (uo, username, imapserver, fnsocket);
    }
  else 
    {
    sm_setproctitle (PROGNAME, true, "%s   (After login BAD. Exiting.)", username);
    ErrorInLogin ();
    }
  
  ExitServer (uo);
  }
    
//MAIN LOOP FOR INTERDAEMON CHILD
int ChildLoop (UserOptions *uo, char *username, char *imapserver, char *fnsocket)
  {
  XSocket *ss;
  int cmd, connected_socket = ERROR_SOCKET, listensock = ERROR_SOCKET, validator, pageid, cmd_counter = 0, numentry;
  TBuffer connid, stlang, abuf, linea, allparms, allparms2, parm1, parm2, parm3, parm4, parm5, attid, contenttype, temp;
  bool FINISHED = false, boo, WasCreatedSentMailFolder = false;
  xulong nummsg, len, cont;
  StringList *SLHeaders, *SLAttach, *SLAttachInfo, *SLAttachHeader, *SLMailboxes;
  SendMSG *sendmsg;
  char *pc;
  Addressbook *ab;
  ULongList *IL;
  
  AlarmInstall ();
  ss = new XSocket ();
  ss->do_server_unix_only_bind (SOCK_STREAM, &listensock, fnsocket);
  LOG ("Login OK IP=%s, user=%s, servercount=%d, totallogins=%ld", ipclient, username, ServerCounter + 1, TotalLogins);
  UnlockAndCloseFile (LockFileDescriptor);
  SLHeaders = new StringList (); 
  SLMailboxes = new StringList ();
  SLAttach = new StringList ();
  SLAttachHeader = new StringList ();
  SLAttachInfo = new StringList ();
  
  while (FINISHED == false)
    {
    //LOOP FOR ACCEPTS OF postman cgis

    sm_setproctitle (PROGNAME, true, "%s   (idle)", username);
    SwitchOnAlarm (uo, Conf->getInt ("main", "timeout", 1200));
    strncpy (BasePath, Conf->getString ("main", "usersbasepath", "/var/postman/users", CMAXBUFFER), CMAXBUFFER);
    connected_socket = accept(listensock, NULL, NULL);
    if (connected_socket < 0) 
      {
      if (errno != EINTR) {perror("accept"); close(listensock); return (ERROR_SOCKET);} else {continue;}
      }
    ss->SetSock (connected_socket);
    ss->Gets (linea, CMAXBUFFER);
    ++cmd_counter;
    html = new HTML (ss);
    html->setMAXMSGSFORPAGE (uo->getMsgsPerIndexPage());    
    cclient->setMAXMSGSFORPAGE (uo->getMsgsPerIndexPage());    
    ini (connid, stlang, stcmd, allparms, allparms2, parm1, parm2, parm3, parm4, parm5);
    sscanf (linea, "%s %d %d %d %s %s %s %[^]]s",      connid, &pageid, &validator, &lastcmd, stlang, ipclient, stcmd, allparms);
    xscanf (linea, connid, &pageid, &validator, &lastcmd, stlang, ipclient, stcmd, parm1, allparms2, NULL, NULL, NULL);
    xscanf (linea, connid, &pageid, &validator, &lastcmd, stlang, ipclient, stcmd, parm1, parm2, parm3, parm4, parm5);
    uo->setPageId (pageid);
    uo->setLastCmd(lastcmd);
    SLHeaders->Clear ();
    SLMailboxes->Clear ();
    SLAttach->Clear ();
    SLAttachHeader->Clear ();
    SLAttachInfo->Clear ();
    cmd = TranslateCmd (stcmd);
    strncpy (lastconnid, connid, CMAXBUFFER);
    
    /*****/ //DEBUG (linea);
    
    sm_setproctitle (PROGNAME, true, "%s   (cmd '%s')", username, stcmd);
    if (!IsGoodValidator (cmd,  validator, uo->getValidator())) {cmd = CMD_INVALIDPAGE;}
    if (cclient->Ping () == false) {LOG("Reconect for IP=%s, user=%s", ipclient, username); cclient->Reconnect();}
    
    uo->addLastStringCmd (linea);
    switch (cmd)
      {
      case (CMD_CHECK_COOKIE): {CheckCookie (uo, parm1, cmd_counter, stlang, connid); break;}
      case (CMD_LOGOUT):
        {
        LOG ("Logout from IP=%s, user=%s", ipclient, username);
        ClearConn ();
        cclient->saveFlagsRCNewsFile ();
        cclient->CloseMailbox ();
        xsnprintf (linea, CMAXBUFFER, "%s?lang=%s", PATH_POSTMAN, stlang);
        ClearCookie (uo);
        html->Redirect(linea);
        FINISHED = true;
        break;
        }
      case (CMD_INVALIDPAGE):
        {
        html->PrintInvalidPage (stlang, username, connid, uo);
        break;
        }        
      case (CMD_MG_DUMPFULLMSG):
        {
        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (long int)1);
        html->dumpPChar (cclient->dumpFullMsg (nummsg));
        break;
        }
      case (CMD_MG_DUMPFULLHEADER):
        {
        TBuffer atmpbuf;
        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (long int)1);
        strncpy (temp, cclient->getFormattedSize(nummsg, atmpbuf), CMAXBUFFER);
        cclient->dumpFullHeader (nummsg, SLHeaders);
        html->PrintFullHeaders(SLHeaders, uo, temp);
        break;
        }                
      case CMD_MG_DISPLAY:
      case CMD_MG_DELDISPLAY:
      case CMD_MG_UNDELDISPLAY:
      case CMD_MG_FG_DELETE:
      case CMD_MG_FG_UNDELETE:
      case CMD_MG_FG_ANSWERED:
      case CMD_MG_FG_UNANSWERED:
      case CMD_MG_FG_FLAGGED:
      case CMD_MG_FG_UNFLAGGED:
      case CMD_MG_FG_SEEN:
      case CMD_MG_FG_UNSEEN:
      case CMD_MG_FG_SEARCHED:
      case CMD_MG_FG_UNSEARCHED:            
        {
        TBuffer aflags, maincontenttype, ato, afrom, areplyto, acc, asubject, adate, asize, atmpbuf;
        long todel, size, ppos;        

        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (xulong)1);
        if (nummsg > (xulong)cclient->getTotalMsgs ()) nummsg = cclient->getTotalMsgs ();        
        if ((cmd == CMD_MG_DELDISPLAY) || (cmd == CMD_MG_UNDELDISPLAY))
          {
          todel = xatoidef (parm1, -100);
          boo = cclient->isLastMsgFromList (todel);
          if (todel != -100)
            {
            ppos = cclient->getPosListFromNumMsg(todel);
            if ((boo == true) && ((unsigned long)uo->getCurrentPosition() == nummsg))
              {
              xsnprintf (abuf, CMAXBUFFER, "%ld", todel);
              }
            else
              {
              todel = cclient->getNumMsgFromPosList(cclient->getPosListFromNumMsg(todel));
              xsnprintf (abuf, CMAXBUFFER, "%ld", todel);
              }
            if (cmd == CMD_MG_DELDISPLAY)
              {
              cont = cclient->setFlags ("\\DELETED", abuf);
              xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_DELETEMSG));
              strcat (abuf, " "); 
              if (boo == false) strcat (abuf, L->get(L_NEXTMSGSHOWN));              
              else strcat (abuf, L->get(L_LASTMSGSHOWN));                                 
              }
            else
              {
              cont = cclient->clearFlags ("\\DELETED", abuf);
              xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNDELETEMSG));              
              strcat (abuf, " "); 
              if (boo == false) strcat (abuf, L->get(L_NEXTMSGSHOWN));              
              else strcat (abuf, L->get(L_LASTMSGSHOWN));                                               
              }
            uo->setStatusLine (abuf, RED);
            }
          }
        else if ((cmd == CMD_MG_FG_DELETE)   || (cmd == CMD_MG_FG_UNDELETE)   || 
                 (cmd == CMD_MG_FG_ANSWERED) || (cmd == CMD_MG_FG_UNANSWERED) || 
                 (cmd == CMD_MG_FG_FLAGGED)  || (cmd == CMD_MG_FG_UNFLAGGED)  || 
                 (cmd == CMD_MG_FG_SEEN)     || (cmd == CMD_MG_FG_UNSEEN)     ||
                 (cmd == CMD_MG_FG_SEARCHED)  || (cmd == CMD_MG_FG_UNSEARCHED)) 
          {
          ppos = cclient->getPosListFromNumMsg(nummsg);
          xsnprintf (abuf, CMAXBUFFER, "%ld", nummsg);
          switch (cmd)
            {
            case CMD_MG_FG_DELETE:  {
                                    cont = cclient->setFlags ("\\DELETED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_DELETEMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_DELETEMSG));}
                                    break;
                                    }
            case CMD_MG_FG_UNDELETE:{
                                    cont = cclient->clearFlags ("\\DELETED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNDELETEMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNDELETEMSG));}
                                    break;
                                    }
            case CMD_MG_FG_ANSWERED:{
                                    cont = cclient->setFlags ("\\ANSWERED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_ANSWEREDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_ANSWEREDMSG));}
                                    break;
                                    }
            case CMD_MG_FG_UNANSWERED:{
                                    cont = cclient->clearFlags ("\\ANSWERED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNANSWEREDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNANSWEREDMSG));}
                                    break;
                                    }
            case CMD_MG_FG_FLAGGED: {
                                    cont = cclient->setFlags ("\\FLAGGED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_FLAGGEDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_FLAGGEDMSG));}
                                    break;
                                    }
            case CMD_MG_FG_UNFLAGGED:{
                                    cont = cclient->clearFlags ("\\FLAGGED", abuf); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNFLAGGEDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNFLAGGEDMSG));}
                                    break;
                                    }
            case CMD_MG_FG_SEARCHED:{ 
                                    cclient->addMsgInSearchList (nummsg);
                                    xsnprintf (abuf, CMAXBUFFER, "%s", L->get(L_SEARCHEDMSG));
                                    break;
                                    }
            case CMD_MG_FG_UNSEARCHED:{
                                    cclient->delMsgInSearchList (nummsg);            
                                    xsnprintf (abuf, CMAXBUFFER, "%s", L->get(L_UNSEARCHEDMSG));
                                    break;
                                    }                                    
            }
          uo->setStatusLine (abuf, RED);  
          }   
                 
        uo->setCurrentPosition (nummsg, cclient->getTotalMsgs ());        
        strncpy (aflags,   cclient->getFlags(nummsg, atmpbuf), CMAXBUFFER);
        if (cclient->getServerProtocol() == PROTOCOL_IMAP)
          {strncpy (ato, cclient->getTo (nummsg), CMAXBUFFER);}
        else
          {strncpy (ato, cclient->getNewsGroup (nummsg), CMAXBUFFER);}
        strncpy (afrom,    cclient->getFrom (nummsg), CMAXBUFFER);
        strncpy (areplyto, cclient->getReplyTo (nummsg), CMAXBUFFER);
        strncpy (acc,      cclient->getCC (nummsg), CMAXBUFFER);        
        strncpy (asubject, cclient->getSubject (nummsg), CMAXBUFFER);
        strncpy (adate,    cclient->getDate (nummsg), CMAXBUFFER);
        strncpy (asize,    cclient->getFormattedSize(nummsg, atmpbuf), CMAXBUFFER);
        cclient->getBodyStructure (nummsg, SLAttach);
        IL = new ULongList ();
        cclient->getMsgIndex (IL);        
        html->PrintMessageHeader (stlang, username, connid, 
                            cclient->getTotalMsgs (), aflags, ato, afrom, areplyto, acc, asubject, adate, asize,
                            SLAttach, uo, IL);        
        strncpy (maincontenttype, cclient->getContentType(nummsg), CMAXBUFFER);
        for (int Y=0; Y<SLAttach->Count(); ++Y)
          {
          sscanf (SLAttach->getString(Y).cstr(), "%s %s %ld %s", attid, contenttype, &size, temp);
          if ((size <= MAXTEXTSIZEATTACHFORDISPLAY) || (Y == 0))
            {
            if (strcmp (contenttype, "TEXT/PLAIN") == 0) 
              {
              if (DISPLAYTXTATTASHTML == 0)
                {               
                if (Y > 0) html->WRITE (L->get(L_NUMATTACHSHOW), attid); html->BR();
                pc = cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true);
                html->BeginPre();
                if (IsNull(pc) == true) {html->WRITE (L->get(MSG_MALFORMED_MESSAGE));}
                else 
                  {
                  /*
                  sprintf (atmpbuf, "<A HREF=\"%s/bd_uudecode/noop/%s/%s/%s/%d/%d/%d/%ld/%s/1/\">%s</A>", 
                           PATH_POSTMAN, stlang, username, connid, uo->getValidator(), uo->getPageId(), 
                           CMD_MG_DISPLAY, uo->getCurrentPosition(), attid, 
                           L->get(L_DISPLAY_UUENCODED));
                  */         
                  sprintf (atmpbuf, "<A HREF=\"%s/bd_uudecode/noop/%s/%s/%s/%d/%d/%d/%ld/%s/", 
                           PATH_POSTMAN, stlang, username, connid, uo->getValidator(), uo->getPageId(), 
                           CMD_MG_DISPLAY, uo->getCurrentPosition(), attid); 
                  ResaltaURLsEnText (pc, ss, uo->getTruncateLengthReadingMsg(), IsUuencodedPChar (pc), atmpbuf);
                  }
                html->EndPre();
                if ((Y != 0) || ((Y == 0) && (SLAttach->Count() > 1))) html->Linea();
                }
              else
                {
                if (Y > 0) html->WRITE (L->get(L_NUMATTACHSHOW), attid); html->BR();
                pc = cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true);
                if (IsNull(pc) == true) {html->WRITE (L->get(MSG_MALFORMED_MESSAGE));}
                else {html->dumpAttachWithoutHead (pc, len);}
                if ((Y != 0) || ((Y == 0) && (SLAttach->Count() > 1))) html->Linea();
                }
              } 
            else if (strcmp (contenttype, "TEXT/HTML") == 0)
              {
              if (Y > 0) html->WRITE (L->get(L_NUMATTACHSHOW), attid); html->BR();
              pc = cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true);
              if (IsNull(pc) == true) {html->WRITE (L->get(MSG_MALFORMED_MESSAGE));}
              else 
                {
                if (DISABLE_ATT_HTML_TOKENS == 1) {DisableHTMLTokens (pc);}
                html->dumpAttachWithoutHead (pc, len);
                }
              if ((Y != 0) || ((Y == 0) && (SLAttach->Count() > 1))) html->Linea();
              }
           //SACO LOS AJUNTOS DE TEXTO QUE NO ENTIENDO POR PANTALLA   
            else if (strstr (contenttype, "TEXT/") != NULL)
              {
              if (Y > 0) html->WRITE (L->get(L_NUMATTACHSHOW), attid); html->BR();
              pc = cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true);
              html->BeginPre();
              if (IsNull(pc) == true) {html->WRITE (L->get(MSG_MALFORMED_MESSAGE));}
              else {html->dumpAttachWithoutHead (pc, len);}
              html->EndPre();
              if ((Y != 0) || ((Y == 0) && (SLAttach->Count() > 1))) html->Linea();              
              }
            }
          } //for
        html->PrintMessageTail (stlang, username, connid, 
                                cclient->getTotalMsgs (), aflags, ato, afrom, areplyto, acc, asubject, adate, asize,
                                SLAttach, uo, IL);        
        html->HTMLEnd(uo->getSkinAtribProp("COLOR_BGTABLECOPYRIGHT"));  
        delete IL;
        break;
        }
      case (CMD_MB_EXPUNGE):
        {
        cont = cclient->getMsgsMarkedDeleted();
        if (cont <= 0) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSPURGED_0));}
        else if (cont == 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSPURGED), cont);}
        else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSPURGED_S), cont);}
        uo->setStatusLine (abuf, RED);
        uo->IncValidator();
        cclient->Expunge ();
        uo->setCurrentPosition (0, cclient->getTotalMsgs ());
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);        
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;
        }
      case (CMD_MB_INDEX): 
        {
        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (long int)0);
        uo->setCurrentPosition (nummsg, cclient->getTotalMsgs ());
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO));
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, WasCreatedSentMailFolder);        
        WasCreatedSentMailFolder = false;
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);               
        break;
        }
      case (CMD_MB_SEARCH):
        {
        int firstnummsgfound, fdsearch;  
                
        if      (strcasecmp (parm1, "fd_subject") == 0) {fdsearch = 1;}
        else if (strcasecmp (parm1, "fd_from")    == 0) {fdsearch = 2;}
        else if (strcasecmp (parm1, "fd_body")    == 0) {fdsearch = 3;}
        else                                            {fdsearch = 0;}
        cont = cclient->searchFolder2 ("ALL", fdsearch, allparms2, NULL, NULL);
        firstnummsgfound = (int)cclient->getNumFirstMsgInSearch();
        if (cont <= 0) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSFOUND_0));}
        else if (cont == 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSFOUND), cont);}
        else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MSGSFOUND_S), cont);}
        uo->setStatusLine (abuf, BLUE);        
        if (firstnummsgfound > 0) {nummsg = firstnummsgfound;}
        else {nummsg = uo->getCurrentPosition ();}        
        uo->setCurrentPosition (nummsg, cclient->getTotalMsgs ());
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO));
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;
        }                
      case (CMD_MB_SORT):
        {
        int sorttype = NO_SORT;
        
        xsnprintf (abuf, CMAXBUFFER, "%s", L->get(MSG_MSGS_SORTED));
        uo->setStatusLine (abuf, RED);
        if      (strcasecmp (parm1, "number")  == 0) {sorttype = NO_SORT;}
        else if (strcasecmp (parm1, "date")    == 0) {if (USE_ARRIVAL_DATE == 1) sorttype = SORTARRIVAL; else sorttype = SORTDATE;}
        else if (strcasecmp (parm1, "from")    == 0) {sorttype = SORTFROM;}
        else if (strcasecmp (parm1, "subject") == 0) {sorttype = SORTSUBJECT;}
        else if (strcasecmp (parm1, "size")    == 0) {sorttype = SORTSIZE;}
        else                                         {sorttype = NO_SORT;}
        cclient->setSortType (sorttype);
        cclient->Sort ();
        uo->setCurrentPosition (1, cclient->getTotalMsgs ());
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;
        }
      case (CMD_MB_SHOW):
        {
        uo->InitStatusLine ();
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        html->PrintMailboxesPage (stlang, cclient->getMailboxName(), username, connid, SLMailboxes, uo);
        break;
        }
      case (CMD_MB_CHANGE):
        {
        uo->InitStatusLine ();
        boo = cclient->ChangeMailbox (parm1);
        if (boo == true)
          {
          uo->IncValidator();
          uo->setStatusLine (L->get(MSG_MAILBOXCHANGED), RED);
          if (cclient->getError()==true) {uo->setStatusLine (cclient->getErrorString(), RED);}
          uo->setCurrentPosition (0, cclient->getTotalMsgs ());
          cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
          uo->setMailboxSize (cclient->getMailboxSize());
          html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
          }
        else 
          {
          TBuffer abuf, abuf2;
          if (cclient->getError() == true) 
            xsnprintf (abuf, CMAXBUFFER, cclient->getErrorString());
          else
            xsnprintf (abuf, CMAXBUFFER, L->get(ERR_INV_MAILBOX), cclient->getFullMailbox(parm1, abuf2));
          html->ErrorPage(abuf);
          }
        break;
        }                        
      case (CMD_MB_CREATE):
        {
        TBuffer atmpbuf;
        strncpy (temp, cclient->expandMailboxName (parm1, atmpbuf), CMAXBUFFER);
        uo->InitStatusLine ();
        boo = cclient->createMailbox (temp);
        if (boo == true)
          {
          uo->setStatusLine (L->get(MSG_MAILBOXCREATED), RED);          
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, true);
          html->PrintMailboxesPage (stlang, cclient->getMailboxName(), username, connid, SLMailboxes, uo);
          }
        else 
          {
          xsnprintf (abuf, CMAXBUFFER, L->get(ERR_INV_MAILBOX), temp);
          html->ErrorPage(abuf);
          }
        break; 
        }                       
      case (CMD_MB_DELETE):
        {
        uo->InitStatusLine ();
        boo = cclient->deleteMailbox (parm1);
        if (boo == true)
          {
          uo->setStatusLine (L->get(MSG_MAILBOXDELETED), RED);          
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, true);
          html->PrintMailboxesPage (stlang, cclient->getMailboxName(), username, connid, SLMailboxes, uo);
          }
        else 
          {
          xsnprintf (abuf, CMAXBUFFER, L->get(ERR_INV_MAILBOX), parm1);
          html->ErrorPage(abuf);
          }
        break; 
        }                       
      case (CMD_MB_RENAME):
        {
        TBuffer atmpbuf;
        strncpy (temp, cclient->expandMailboxName (parm2, atmpbuf), CMAXBUFFER);
        uo->InitStatusLine ();
        boo = cclient->renameMailbox (parm1, temp);
        if (boo == true)
          {
          uo->setStatusLine (L->get(MSG_MAILBOXRENAMED), RED);          
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, true);
          html->PrintMailboxesPage (stlang, cclient->getMailboxName(), username, connid, SLMailboxes, uo);
          }
        else 
          {
          xsnprintf (abuf, CMAXBUFFER, L->get(ERR_INV_MAILBOX), temp);
          html->ErrorPage(abuf);
          }
        break;
        }   
      case (CMD_MB_DUMP):
        {                    
        uo->InitStatusLine ();
        html->Head ("text/plain", "", "");
        for (int i = 1; i <= cclient->getTotalMsgs(); ++i)
          {
          time_t now; now = time (NULL);
          html->pcWRITE ("From Postman ");
          html->pcWRITE (ctime(&now));
          html->pcWRITE (cclient->dumpFullMsg (i));
          html->pcWRITE ("\r\n");
          }
        break;
        }
      case CMD_FG_DELETE:
      case CMD_FG_UNDELETE:
      case CMD_FG_ANSWERED:
      case CMD_FG_UNANSWERED:
      case CMD_FG_FLAGGED:
      case CMD_FG_UNFLAGGED:
      case CMD_FG_SEEN:
      case CMD_FG_UNSEEN:
        {
        if (strlen(allparms) == 0) {uo->setStatusLine (L->get(MSG_NO_ENTRIES_MARKED), RED);}
        else
          {
          switch (cmd)
            {
            case CMD_FG_DELETE:     {
                                    cont = cclient->setFlags ("\\DELETED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_DELETEMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_DELETEMSG));}
                                    break;
                                    }
            case CMD_FG_UNDELETE:   {
                                    cont = cclient->clearFlags ("\\DELETED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNDELETEMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNDELETEMSG));}
                                    break;
                                    }
            case CMD_FG_ANSWERED:   {
                                    cont = cclient->setFlags ("\\ANSWERED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_ANSWEREDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_ANSWEREDMSG));}
                                    break;
                                    }
            case CMD_FG_UNANSWERED: {
                                    cont = cclient->clearFlags ("\\ANSWERED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNANSWEREDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNANSWEREDMSG));}
                                    break;
                                    }
            case CMD_FG_FLAGGED:    {
                                    cont = cclient->setFlags ("\\FLAGGED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_FLAGGEDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_FLAGGEDMSG));}
                                    break;
                                    }
            case CMD_FG_UNFLAGGED:  {
                                    cont = cclient->clearFlags ("\\FLAGGED", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNFLAGGEDMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNFLAGGEDMSG));}
                                    break;
                                    }
            case CMD_FG_SEEN:       {
                                    cont = cclient->setFlags ("\\SEEN", allparms); 
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_SEENMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_SEENMSG));}
                                    break;
                                    }
            case CMD_FG_UNSEEN:     {
                                    cont = cclient->clearFlags ("\\SEEN", allparms);
                                    if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED), cont, L->get(L_UNSEENMSG));}
                                    else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_FLAGS_CHANGED_S), cont, L->get(L_UNSEENMSG));}
                                    break;
                                    }
            }
          uo->setStatusLine (abuf, RED);  
          }
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;
        }
      case CMD_MG_MOVE:
        {
        if (strlen(allparms2) == 0) {uo->setStatusLine (L->get(MSG_NO_ENTRIES_MARKED), RED);}
        else
          {  
          cont = cclient->moveMessages (parm1, allparms2);
          if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MOVEDMESSAGES), cont, parm1);}
          else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_MOVEDMESSAGES_S), cont, parm1);}
          uo->setStatusLine (abuf, RED);
          }
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break; 
        } 
      case CMD_MG_COPY:
        {
        if (strlen(allparms2) == 0) {uo->setStatusLine (L->get(MSG_NO_ENTRIES_MARKED), RED);}
        else
          {                
          cont = cclient->copyMessages (parm1, allparms2);
          if (cont <= 1) {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_COPYMESSAGES), cont, parm1);}
          else {xsnprintf (abuf, CMAXBUFFER, L->get(MSG_COPYMESSAGES_S), cont, parm1);}
          uo->setStatusLine (abuf, RED);
          }
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO)); 
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;
        }
      case CMD_CONF_ACT:
        {
        /* DIWovBTkom694Qm19aq80Jy eng conf_act mb_delete alternativo*/
        uo->InitStatusLine ();
        html->PrintConfirmPage (parm1, stlang, parm2, username, connid, uo->getValidator(), uo->getPageId(), uo);
        break;   
        }
      case (CMD_CM_SAVE):
        {        
        CopyFileOverWrite (parm1, uo->getfnSavedMsg());
        DeleteFile (parm1);
        uo->InitStatusLine ();
        html->PrintComposeMsgPage (CMD_CM, CMD_CM_SAVE, stlang, username, connid, 
                                   NULL, NULL, NULL, NULL, 
                                   NULL, NULL, NULL,
                                   NULL,
                                   uo);          
        break; 
        }        
      case (CMD_CM_CLEANALL):
        {
        if (FileExists (uo->getfnSavedMsg()) == true) DeleteFile (uo->getfnSavedMsg());
        uo->delAllAttachs();
        uo->InitStatusLine ();
        html->PrintComposeMsgPage (CMD_CM, CMD_CM_CLEANALL, stlang, username, connid, 
                                   NULL, NULL, NULL, NULL, 
                                   NULL, NULL, NULL,
                                   NULL,
                                   uo);          
      
        break;
        }
      case (CMD_CM):
        {
        uo->setNumMsgToReply (-1);
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        html->PrintComposeMsgPage (CMD_CM, CMD_CM, stlang, username, connid, 
                                   NULL, NULL, NULL, NULL, 
                                   NULL, NULL, NULL,
                                   NULL,
                                   uo);
        break;
        }
      case (CMD_CMNNTP):
        {
        uo->setNumMsgToReply (-1);
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        html->PrintComposeMsgPage (CMD_CMNNTP, CMD_CM, stlang, username, connid, 
                                   NULL, NULL, NULL, NULL, 
                                   NULL, NULL, NULL,
                                   NULL,
                                   uo);
        break;
        }        
      case (CMD_CM_REPLY):
        {
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        nummsg = xatoldef (parm1, (long int)1);  
        uo->setNumMsgToReply ((int)nummsg);        
        cclient->getBodyStructure (nummsg, SLAttach);
        sscanf (SLAttach->getString(0).cstr(), "%s %s %s", attid, contenttype, temp);
        //I need that xtrdup for nntp. DO NOT DELETE.
        pc = xstrdup (cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true));
        html->PrintComposeMsgPage (CMD_CM_REPLY, CMD_CM_REPLY, stlang, username, connid, 
                                   cclient->getTo (nummsg), NULL, NULL, cclient->getSubject (nummsg), 
                                   cclient->getReplyTo(nummsg), cclient->getFrom(nummsg), cclient->getDate(nummsg), 
                                   pc,
                                   uo);                                   
                                   
        delete pc;
        break;  
        }      
      case (CMD_CM_REPLYGROUP):
        {
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        nummsg = xatoldef (parm1, (long int)1);  
        uo->setNumMsgToReply ((int)nummsg);        
        cclient->getBodyStructure (nummsg, SLAttach);
        sscanf (SLAttach->getString(0).cstr(), "%s %s %s", attid, contenttype, temp);
        //I need that xtrdup for nntp. DO NOT DELETE.
        pc = xstrdup (cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true));
        html->PrintComposeMsgPage (CMD_CM_REPLYGROUP, CMD_CM_REPLYGROUP, stlang, username, connid, 
                                   cclient->getTo (nummsg), NULL, NULL, cclient->getSubject (nummsg), 
                                   cclient->getReplyTo(nummsg), cclient->getFrom(nummsg), cclient->getDate(nummsg), 
                                   pc,
                                   uo);
        delete pc;
        break;  
        }              
      case (CMD_CM_REPLYALL):
        {
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        nummsg = xatoldef (parm1, (long int)1);  
        uo->setNumMsgToReply ((int)nummsg);        
        cclient->getBodyStructure (nummsg, SLAttach);
        sscanf (SLAttach->getString(0).cstr(), "%s %s %s", attid, contenttype, temp);
        pc = xstrdup (cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true));
        html->PrintComposeMsgPage (CMD_CM_REPLYALL, CMD_CM_REPLYALL, stlang, username, connid, 
                                   cclient->getTo (nummsg), cclient->getCC (nummsg), NULL, cclient->getSubject (nummsg), 
                                   cclient->getReplyTo(nummsg), cclient->getFrom(nummsg), cclient->getDate(nummsg), 
                                   pc,
                                   uo);
        delete pc;
        break;  
        }                      
      case (CMD_CM_FORWARD):
        {
        xulong size;
        int Y, num;
        TBuffer ftotname, fname, bas, info;
        
        uo->setStatusLine (L->get(MSG_COMPOSE_TIMEOUT), uo->getSkinAtribProp("COLOR_FONT_STATUSLINE"));
        uo->setNumMsgToReply (-1);
        nummsg = xatoldef (parm1, (long int)1);
        
        if ((uo->getForwAttachs () == false) || (uo->getLastCmd () != CMD_MG_DISPLAY))
          {//*** NO INSERT THE ATTACHS IN THE FORWARD
          cclient->getBodyStructure (nummsg, SLAttach);
          sscanf (SLAttach->getString(0).cstr(), "%s %s %s", attid, contenttype, temp);
          pc = xstrdup (cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true));
          html->PrintComposeMsgPage (CMD_CM_FORWARD, CMD_CM_FORWARD, stlang, username, connid, 
                                     cclient->getTo (nummsg), NULL, NULL, cclient->getSubject (nummsg), 
                                     cclient->getReplyTo(nummsg), cclient->getFrom(nummsg), cclient->getDate(nummsg),
                                     pc,
                                     uo);        
          }
        else
          {//*** SI INSERT
          cclient->getBodyStructure (nummsg, SLAttach);
          Y = 0; num = 0;
          while (Y < SLAttach->Count())
            {
            //2 IMAGE/GIF 154387 bytes, "un comentario", "logo.gif"
            initStr (temp); initStr (info);
            sscanf (SLAttach->getString(Y).cstr(), "%s %s %ld %s %[^]]s", attid, contenttype, &size, temp, info);
            if (strchr (attid, '.') == NULL)
              {
              if ((Y > 0) || ((Y == 0) && (strcmp (contenttype, "TEXT/PLAIN") != 0)))
                {
                ++num;                
                char *filename = get_filename(SLAttach->getString(Y).cstr());
                initStr(fname, filename);
                free(filename);                                
                strncpy (ftotname, (char *)uo->getAttachDir(), CMAXBUFFER); 
                strcat (ftotname, "/"); strcat (ftotname, fname);
                cclient->getAttBodyToFile (nummsg, attid, SLAttachInfo, &len, ftotname, true);
                xsnprintf (bas, CMAXBUFFER, "%s %s", contenttype, info);
                uo->addFileNameAtach (fname, bas);
                }
              }               
            ++Y;
            }
          sscanf (SLAttach->getString(0).cstr(), "%s %s %s", attid, contenttype, temp);
          pc = xstrdup (cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true));
          html->PrintComposeMsgPage (CMD_CM_FORWARD, CMD_CM_FORWARD, stlang, username, connid, 
                                     cclient->getTo (nummsg), NULL, NULL, cclient->getSubject (nummsg), 
                                     cclient->getReplyTo(nummsg), cclient->getFrom(nummsg), cclient->getDate(nummsg),
                                     pc,
                                     uo);
          }
        delete pc;
        break;   
        }             
      case (CMD_CM_SEND):
        {
        TNumber anum;
        CopyFileOverWrite (parm1, uo->getfnSavedMsg());
        int numattachs = xatoidef (parm2, 0);
        if (numattachs != uo->numAttachs())
          {
          html->ErrorPageBack(stlang, username, connid, uo, L->get(MSG_ERR_RELOAD_ATTACHS), 2);
          }
        else
          {
          uo->InitStatusLine ();
          sendmsg = new SendMSG (uo->getSMTPServer(), parm1, uo, uo->getImapServer(), cclient);
          WasCreatedSentMailFolder = false;
          boo = sendmsg->sendSMTP(temp, uo->getSentMailName(), &WasCreatedSentMailFolder, MAILCAP);
          if (boo == true) 
            {
            int anummsg;
            anummsg  = uo->getNumMsgToReply ();
            if (anummsg > -1) {cclient->setFlags ("\\ANSWERED", xitoa(anummsg, anum));}
            if (strcmp (temp, "") == 0)
              {
              html->OkPage (stlang, username, connid, uo, L->get(MSG_SEND_OK));
              }
            else
              {
              //SENT, BUT CAN NOT COPY IN SENTMAIL
              html->OkPage (stlang, username, connid, uo, L->get(MSG_ERR_SAVE_SENTMAIL));
              }
            sendmsg->DelAttachs ();  
            }
          else 
            {
            html->ErrorPage (temp);
            }
          delete sendmsg;
          }
        DeleteFile (parm1);
        break;
        }
      case (CMD_CM_SENDNNTP):
        {
        TNumber anum;
        CopyFileOverWrite (parm1, uo->getfnSavedMsg());
        int numattachs = xatoidef (parm2, 0);
        if (numattachs != uo->numAttachs())
          {
          html->ErrorPageBack(stlang, username, connid, uo, L->get(MSG_ERR_RELOAD_ATTACHS), 2);
          }
        else
          {
          uo->InitStatusLine ();
          sendmsg = new SendMSG (uo->getNNTPServer(), parm1, uo, uo->getImapServer(), cclient);
          WasCreatedSentMailFolder = false;
          boo = sendmsg->sendNNTP(temp, uo->getSentMailName(), &WasCreatedSentMailFolder, MAILCAP);
          if (boo == true) 
            {
            int anummsg;
            anummsg  = uo->getNumMsgToReply ();
            if (anummsg > -1) {cclient->setFlags ("\\ANSWERED", xitoa(anummsg, anum));}
            if (strcmp (temp, "") == 0)
              {
              html->OkPage (stlang, username, connid, uo, L->get(MSG_SEND_OK));
              }
            else
              {
              //SENT, BUT CAN NOT COPY IN SENTMAIL
              html->OkPage (stlang, username, connid, uo, L->get(MSG_ERR_SAVE_SENTMAIL));
              }
            sendmsg->DelAttachs ();  
            }
          else 
            {
            html->ErrorPage (temp);
            }
          delete sendmsg;
          }
        DeleteFile (parm1);
        break;
        }        
      case (CMD_CM_CANCEL):
        {
        uo->setStatusLine (L->get(MSG_CANCELMSG), RED);
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO));
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break;  
        }      
      case (CMD_OP):
        {
        uo->setStatusLine (L->get(MSG_SHOWOPTIONS), RED);
        html->PrintShowOptions (stlang, username, connid, uo);
        break;
        }              
      case (CMD_OP_SAVE):
        {
        boo = uo->SaveOptions(parm1);
        if (boo == true) 
          {
          uo->setStatusLine (L->get(MSG_SAVEOPTIONS), RED); uo->LoadOptions();
          if (ALLOW_NNTP == 1)
            {
            cclient->setNNTPServer (uo->getNNTPServer());
            CN->setNNTPserver (uo->getNNTPServer());
            }
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
          }
        else uo->setStatusLine (L->get(MSG_ERR_SAVEOPTIONS), RED);     
        html->PrintShowOptions (stlang, username, connid, uo);
        break;  
        }                    
      case (CMD_OP_CANCEL):
        {
        uo->setStatusLine (L->get(MSG_CANCELOPTIONS), RED);
        cclient->getHeaderList (uo->getBeginPage(), SLHeaders, username, uo->getMailDomain(), L->get (L_TO));
        cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
        uo->setMailboxSize (cclient->getMailboxSize());        
        html->PrintIndexPage (stlang, cclient->getMailboxName(), username, connid, SLHeaders, SLMailboxes, uo);
        break; 
        }       
      case (CMD_AT_DUMP):
        {
        TBuffer atmpbuf;
        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (long int)1);
        //parm2 es el attid, un pchar del tipo "12.3.4"
        //    si es del tipo cid:03d501bf68d6$5465a1e0$0401a8c0@Pc3.ncsa.es
        //    debo convertirlo al anterior
        if (NCstrstr(parm2, "CID:") > 0)
          {
          strncpy (temp, cclient->getAttIdFromCID (nummsg, parm2, atmpbuf), CMAXBUFFER);
          } 
        else 
          {
          strncpy (temp, parm2, CMAXBUFFER);  
          }
        cclient->getAttHeaderSL(nummsg, temp, SLAttachHeader);
        if (cclient->getAttSize (nummsg, temp, SLAttach) > UNHANDLESIZEATTACH)
          {
          html->ErrorPage(L->get(ERR_UNHANDLESIZEATTACH));
          }
        else
          {          
          pc = cclient->getAttBody (nummsg, temp, SLAttach, &len, true);
          if ((pc != NULL) && (len > 0))
            {
            html->dumpRawAttachSL (SLAttachHeader, pc, len);
            }
          else
            {
            html->EmulatePageNotFound ();
            }
          }    
        break;  
        }                      
      case (CMD_AT_SHOW):
        {
        CopyFileOverWrite (parm1, uo->getfnSavedMsg());
        DeleteFile (parm1);  
        uo->setStatusLine (L->get(MSG_SHOWATTACHS), RED);
        html->PrintShowAttachs (stlang, username, connid, uo);
        break;  
        }            
      case (CMD_AT_SAVE):
        {
        int attssaved;
        xsnprintf (temp, CMAXBUFFER, "%s \"\", \"%s\"", getMailcapTipoFromFilename(allparms, MAILCAP), allparms);
        attssaved = uo->addFileNameAtach (allparms, temp);
        if (attssaved > 0) uo->setStatusLine (L->get(MSG_SAVEATTACHS), RED);
        else uo->setStatusLine (L->get(MSG_NOATTACHSSAVED), RED);
        html->PrintShowAttachs (stlang, username, connid, uo);
        break;
        }   
      case (CMD_AT_ERASE):
        {
        int totdel;
        totdel = uo->delFileNameAtach (allparms);
        if (totdel > 0) uo->setStatusLine (L->get(MSG_ERASEATTACHS), RED);
        else uo->setStatusLine (L->get(MSG_NOATTACHSERASED), RED);
        html->PrintShowAttachs (stlang, username, connid, uo);
        break; 
        } 
      case (CMD_BD_UUDECODE):
        {
        TBuffer TrueName, fntmpin, fntmpout, CT;
        uo->InitStatusLine ();
        nummsg = xatoldef (parm1, (long int)1);
        int numatt = xatoldef (parm2, (long int)1);
        cclient->getBodyStructure (nummsg, SLAttach);
        pc = cclient->getAttBody (nummsg, attid, SLAttachInfo, &len, true);
        MakeTmpName (fntmpin, CMAXBUFFER);
        MakeTmpName (fntmpout, CMAXBUFFER);
        if (SavePchar (pc, len, fntmpin) == false) 
          {   
          html->ErrorPage (L->get(ERR_UUDECODING));
          } 
        else if (UudecodeFiles (fntmpin, fntmpout, TrueName, numatt) == false) 
          {
          html->ErrorPage (L->get(ERR_UUDECODING));
          } 
        else
          {
          getMailcapTipoFromFilename (TrueName, MAILCAP);
          strncpy (CT, getMailcapTipoFromFilename (TrueName, MAILCAP), CMAXBUFFER);
          html->dumpFile (fntmpout, CT);
          }
        DeleteFile (fntmpin);
        DeleteFile (fntmpout);
        break;  
        }                              
      case (CMD_AB_SHOWALL):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        html->PrintAddBook_Show (cmd, stlang, username, connid, uo, ab, -1);
        delete ab;
        break;break; 
        }  
      case (CMD_AB_DUMP):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        uo->setStatusLine (L->get(MSG_ADDBOOK), RED);
        html->PrintAddBook_Dump (cmd, stlang, username, connid, uo, ab, -1);
        delete ab;
        break; 
        }          
      case (CMD_AB_SORT):
      case (CMD_AB_FROM_CM):
        {
        int sortby = -1;
        
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        uo->setStatusLine (L->get(MSG_ADDBOOK), RED);
        if (cmd == CMD_AB_SORT) {sortby = xatoidef (parm1, -1);}    
        if (cmd == CMD_AB_FROM_CM) 
          {
          CopyFileOverWrite (parm1, uo->getfnSavedMsg());
          DeleteFile (parm1);  
          } 
        html->PrintAddBook_Show (cmd, stlang, username, connid, uo, ab, sortby);
        delete ab;
        break; 
        }
      case (CMD_AB_TAKEADDRESS):
        {
        nummsg = xatoldef (parm1, (long int)1);
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        uo->setStatusLine (L->get(MSG_AB_EDITENTRY), RED);
        numentry = ab->AddNewEntry(cclient->getFrom (nummsg));      
        html->PrintAddBook_EditEntry (cmd, stlang, username, connid, uo, ab, numentry);
        delete ab;
        break; 
        }
      case (CMD_AB_EDITENTRY):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        uo->setStatusLine (L->get(MSG_AB_EDITENTRY), RED);
        numentry = xatoidef (parm1, (int)0);      
        html->PrintAddBook_EditEntry (cmd, stlang, username, connid, uo, ab, numentry);
        delete ab;
        break; 
        }
      case (CMD_AB_NEWENTRY):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        uo->setStatusLine (L->get(MSG_AB_EDITENTRY), RED);
        html->PrintAddBook_EditEntry (cmd, stlang, username, connid, uo, ab, -1);
        delete ab;
        break; 
        }                
      case (CMD_AB_DELEENTRIES):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);        
        if (strlen(allparms) == 0) {uo->setStatusLine (L->get(MSG_NO_ENTRIES_MARKED), RED);}
        else if (strcmp (allparms, "-1") != 0)
          {
          boo = ab->DeleteEntries (allparms);
          if (boo) {uo->setStatusLine (L->get(MSG_AB_DELEENTRY), RED);}
          else {uo->setStatusLine (L->get(ERR_AB_DELEENTRY), RED);}
          }  
        html->PrintAddBook_Show (cmd, stlang, username, connid, uo, ab, -1);
        delete ab;
        break;        
        }        
      case (CMD_AB_SAVEENTRY):
        {
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        numentry = xatoidef (parm1, (int)-1);
        if (numentry == -1) {numentry = ab->AddNewEntry();}
        ab->UpdateEntryFromFile (numentry, parm2);
        uo->setStatusLine (L->get(MSG_AB_SAVEENTRY), RED);
        delete ab;
        xsnprintf (temp, CMAXBUFFER, "%s", uo->getLastStringCmd (CMD_AB_SAVEENTRY, stlang, connid, false));
        html->Redirect (temp);
        break; 
        }
      case (CMD_AB_ADDFIELDS):
        {
        int n1 = 0, n2 = 0, n3 = 0;
        bool b1 = false, b2 = false, b3 = false;
        TBuffer lineaTO, lineaCC, lineaBCC, tmpbuf;
        
        ab = new Addressbook (ReturnUserAddressbookFileName(BasePath, username, imapserver, temp), uo);
        if (FileExists (uo->getfnSavedMsg()) == false) {CreaFileSavedMsg(uo->getfnSavedMsg());}
        strncpy (lineaTO,  getLineFile(parm1, 1, 1, tmpbuf), CMAXBUFFER);
        strncpy (lineaCC,  getLineFile(parm1, 2, 2, tmpbuf), CMAXBUFFER);        
        strncpy (lineaBCC, getLineFile(parm1, 3, 3, tmpbuf), CMAXBUFFER);   
        if (strcmp(lineaTO, " ") != 0)
          {     
          b1 = ConcatLineToLineOfFile (uo->getfnSavedMsg(), 2, ab->getMultiAddresses(lineaTO),  ",", MAXLENGTH_TO); 
          n1 = CountTokensEnString (ab->getMultiAddresses(lineaTO), ',');
          }
        if (strcmp(lineaCC, " ") != 0)  
          {
          b2 = ConcatLineToLineOfFile (uo->getfnSavedMsg(), 3, ab->getMultiAddresses(lineaCC),  ",", MAXLENGTH_TO);         
          n2 = CountTokensEnString (ab->getMultiAddresses(lineaCC), ',');
          }
        if (strcmp(lineaBCC, " ") != 0)
          {
          b3 = ConcatLineToLineOfFile (uo->getfnSavedMsg(), 4, ab->getMultiAddresses(lineaBCC), ",", MAXLENGTH_TO);         
          n3 = CountTokensEnString (ab->getMultiAddresses(lineaBCC), ',');
          }
        if ((b1 == true) || (b2 == true) || (b3 == true))
          {
          uo->setStatusLine (L->get(MSG_ADDRESS_LINE_CUT), RED);
          }
        else if ((strcmp(lineaTO, " ") == 0) && (strcmp(lineaCC, " ") == 0) && (strcmp(lineaBCC, " ") == 0))
          {
          uo->setStatusLine (L->get(MSG_NO_ENTRIES_MARKED), RED);
          }  
        else
          {
          xsnprintf (temp, CMAXBUFFER, L->get(MSG_AB_ADDFIELDS), n1 + n2 + n3);
          uo->setStatusLine (temp, RED);
          }  
        DeleteFile (parm1);  
        html->PrintAddBook_Show (cmd, stlang, username, connid, uo, ab, -1);  
        delete ab;
        break;
        }        
      case (CMD_OP_NNTP):
        {
        CN->setNNTPserver (uo->getNNTPServer());
        CN->retrieveAllGroupsList ();
        CN->retrievePartialGroupsList ("");
        uo->setStatusLine (L->get(MSG_MANAGE_NNTPGROUPS), RED);
        html->PrintShowNNTPManageGroups (stlang, username, connid, uo, CN, NULL);
        break;
        }              
      case (CMD_OP_NNTP_SEARCH):
        {
        int found;          
        CN->setNNTPserver (uo->getNNTPServer());
        CN->retrieveAllGroupsList ();
        found = CN->retrievePartialGroupsList (parm1);
        if (found < MAX_NNTP_GROUPS_FOUND)
          xsnprintf (temp, CMAXBUFFER, L->get(MSG_NNTPGROUPS_FOUND), found);
        else
          xsnprintf (temp, CMAXBUFFER, L->get(MSG_MAX_NNTPGROUPS_FOUND), found);
        uo->setStatusLine (temp, RED);
        html->PrintShowNNTPManageGroups (stlang, username, connid, uo, CN, parm1);
        break; 
        }                                                
      case (CMD_OP_NNTP_SAVE):
        {
        boo = uo->SaveNNTPOptions(parm1);
        if (boo == true) 
          {
          uo->setStatusLine (L->get(MSG_SAVEOPTIONS), RED); uo->LoadOptions();
          cclient->setNNTPServer (uo->getNNTPServer());
          CN->setNNTPserver (uo->getNNTPServer());
          cclient->getMailboxes (SLMailboxes, uo->SLSubscribedGroups, false);
          }
        else uo->setStatusLine (L->get(MSG_ERR_SAVEOPTIONS), RED);
        CN->retrieveAllGroupsList ();
        CN->retrievePartialGroupsList (parm1);
        html->PrintShowNNTPManageGroups (stlang, username, connid, uo, CN, NULL);
        DeleteFile (parm1);
        break;  
        }                    
      case (CMD_OP_NNTP_ADDSELECTED):
        {
        TBuffer tmpbuf;
        int added;
        char *linea;
        StringList *SLTemp; 
        SLTemp = new StringList ();
        linea = getLineFile(parm1, 1, 1, tmpbuf);
        SeparaTokensEnStringList (linea, SP, SLTemp);
        added = MergeStringList (uo->SLSubscribedGroups, SLTemp, true);
        uo->SaveSubscribedGroups();
        CN->setNNTPserver (uo->getNNTPServer());
        CN->retrieveAllGroupsList ();
        xsnprintf (temp, CMAXBUFFER, L->get(MSG_NNTPGROUPS_ADDED), added);
        uo->setStatusLine (temp, RED);
        html->PrintShowNNTPManageGroups (stlang, username, connid, uo, CN, NULL);
        DeleteFile (parm1);
        delete SLTemp;
        break; 
        } 
      case (CMD_NW_LIST_DUMP):
        {
        uo->InitStatusLine ();
        CN->setNNTPserver (uo->getNNTPServer());
        CN->retrieveAllGroupsList ();
        html->dumpSL (CN->SLAllGroups, true);
        break;
        }                                                                                      
      case (CMD_DEBUG):
        {
        uo->setStatusLine (L->get(MSG_DEBUG), RED);
        html->PrintDebugPage (stlang, username, connid, uo);
        break; 
        }
      default: 
        {     
        cclient->CloseMailbox ();
        html->Redirect(PATH_POSTMAN);
        FINISHED = true;
        break; 
        }         
      }
    close (connected_socket);  
    }  //END OF LOOP
    
  delete SLAttachInfo;  
  delete SLAttachHeader;
  delete SLAttach;  
  delete SLMailboxes;
  delete SLHeaders;
  ss->Close ();
  DeleteFile (fnsocket);
  return 0;             
  } 

void procsignals (int signal_type)
  {
  TBuffer abuf, temp;  
  bool boo;
  time_t now;
  
  switch (signal_type)
    {
    /* TIMEOUT. Alarm signal. Kill the child server. */     
    case SIGALRM:
         LOG ("Timeout for IP=%s, user=%s", ipclient, username);
         if (ALLOW_SEVERAL_CONNS == 0) {DeleteFile (ReturnUserLockFileName (username, temp));}
         strncpy (abuf, make_session_socket (username, lastconnid, temp), CMAXBUFFER);
         DeleteFile (abuf);
         exit (0);    
         break;              
    /* ZOMBIES. This waits for all children, so that they don't become zombies. */
    case SIGCHLD:    
         int pid, status;
         --ServerCounter;
         while ((pid = wait3(&status, WNOHANG, NULL)) > 0);    
         break;
    /* SIGHUP. Reload configuration. */         
    case SIGHUP:
         boo = Conf->reload();
         if (boo == true) LOG ("Reloading configuration. Successfull.");
         else LOG ("Error reloading configuration.");    
         ServerCounter = getNumChilds ();
         LOG ("    ServerCounter is now=%ld", ServerCounter);
         break;
    /* SIGTERM. Normal kill sent to daemon*/     
    case SIGTERM:
         if (getpid () == MainServerPid)
           {
           LOG ("Main server is down.");
           }
         else
           {
           LOG ("SIGTERM received for IP=%s, user=%s", ipclient, username);
           if (ALLOW_SEVERAL_CONNS == 0) {DeleteFile (ReturnUserLockFileName (username, temp));}
           strncpy (abuf, make_session_socket (username, lastconnid, temp), CMAXBUFFER);
           DeleteFile (abuf);
           }
         exit (0);      
         break;
    /*SIGSEGV. Invalid memory reference*/     
    case SIGSEGV:
         now = time (NULL);
         fprintf (stderr, "Received signal SIGSEGV for IP=%s, user=%s, date=%s", ipclient, username, ctime(&now));
         exit (1);
         break;
    /*SIGBUS. Bus error (bad memory access)*/
    case SIGBUS:
         now = time (NULL);
         fprintf (stderr, "Received signal SIGBUS for IP=%s, user=%s, date=%s", ipclient, username, ctime(&now));
         exit (1);
         break;                  
    /* SIGUSR1. Program signal*/     
    case SIGUSR1:
         if (getpid () == MainServerPid)
           {
           LOG ("Main server is down.");
           }
         else
           {
           LOG ("SIGUSR1 received for IP=%s, user=%s", ipclient, username);
           strncpy (abuf, make_session_socket (username, lastconnid, temp), CMAXBUFFER);
           DeleteFile (abuf);
           }
         exit (0);  
         break;
    /* SIGUSR2. Program signal*/     
    case SIGUSR2:
         ++ServerCounter;
         ++TotalLogins;  
         break;         
    }    
  }

void SetHandlers (void)
  {
  struct sigaction act1, oldact1, act2, oldact2, act3, oldact3, act4, oldact4, act5, oldact5, act6, oldact6, act7, oldact7;
  sigemptyset(&act1.sa_mask); act1.sa_flags = 0; act1.sa_handler = procsignals; sigaction(SIGCHLD, &act1, &oldact1); 
  sigemptyset(&act2.sa_mask); act2.sa_flags = 0; act2.sa_handler = procsignals; sigaction(SIGHUP,  &act2, &oldact2); 
  sigemptyset(&act3.sa_mask); act3.sa_flags = 0; act3.sa_handler = procsignals; sigaction(SIGTERM, &act3, &oldact3);   
  sigemptyset(&act4.sa_mask); act4.sa_flags = 0; act4.sa_handler = procsignals; sigaction(SIGUSR1, &act4, &oldact4);     
  sigemptyset(&act5.sa_mask); act5.sa_flags = 0; act5.sa_handler = procsignals; sigaction(SIGSEGV, &act5, &oldact5);       
  sigemptyset(&act6.sa_mask); act6.sa_flags = 0; act6.sa_handler = procsignals; sigaction(SIGBUS,  &act6, &oldact6);       
  sigemptyset(&act7.sa_mask); act7.sa_flags = 0; act7.sa_handler = procsignals; sigaction(SIGUSR2, &act7, &oldact7);
  }

void initServer_beforeFork(int argc, char **argv, char **envp, int *cs)
  {    
  int listensock = -1;
  
  MainServerPid = WritePidFile ();  
  if (argc != 2) {printf ("Sintax is: interdaemon fileconfname\n"); exit (1);} 
  Conf = new Configuration (argv[1], true);
  if (!Conf->ConfRead ()) {printf ("Error: %s do not exists.\n", argv[1]); exit (1);}      
  umask (UMASK);    
  OPENLOG();
  LOG ("Main server is started.");    
  Randomize ();
  serversocket = new XSocket ();
  serversocket->ignore_pipe();
  SetHandlers(); 
  *cs = serversocket->do_server_unix (SOCK_STREAM, &listensock);
  }

void initServer_afterFork (int argc, char **argv, char **envp, int cs)
  {
  initsetproctitle(argc, argv, envp);
  //Envio una senyal de login al padre para que incremente los contadores necesarios
  kill (MainServerPid, SIGUSR2);  
  Randomize ();
  serversocket->SetSock (cs);
  }

void CheckServerCount (char *astlang)
  {
  if (ServerCounter > Conf->getInt ("main", "maxconnallowed", 50))
    {
    LOG ("Max connections allowed %d reached.", ServerCounter);
    html->MsgPage(L->get (MSG_TOO_MAX_CONN), astlang);
    fflush (stdout); exit (0);
    }  
  }  
                                                                  
void CheckParams (int cmd, char *connid, char *pw)
  {
  if ((cmd != CMD_LOGIN) || (strcmp(connid, "-1") != 0) || (strcmp(username, "") == 0) || (strcmp(pw, "") == 0))
    {
    html->ErrorPage(L->get (ERR_INV_PARMS)); 
    fflush (stdout); exit (0);
    }
  }      

void CheckAllowServeralConns (void)
  {      
  if (ALLOW_SEVERAL_CONNS == 0)  //NO simultaneous connections
    {            
    long LOCKTIME, NOW, INTERVAL;
    TBuffer temp;
    LOCKTIME = GetTimeFromLockFileOpening (ReturnUserLockFileName (username, temp));      
    NOW = time (NULL);
    INTERVAL = NOW - LOCKTIME;
    if ((INTERVAL <= MAX_TIME_BETWEEN_CONNS) && (LOCKTIME != -1))
      {
      --ServerCounter;
      LOG ("Time minimum between connections. User '%s'. Exiting.", username);
      html->ErrorPage (L->get(MSG_MIN_TIMEBETWCONNS)); 
      exit(0);        
      }
    }       
  }
  
void ClearConn (void)
  {
  TBuffer temp;
  if (ALLOW_SEVERAL_CONNS == 0) {DeleteFile (ReturnUserLockFileName (username, temp));}
  }
  
//ALARM INSTALL, FOR TIMEOUT
void AlarmInstall (void)
  {
  struct sigaction actx, oldactx;
  if (USETIMEOUT == 1)
    {
    sigemptyset(&actx.sa_mask); actx.sa_flags = 0; actx.sa_handler = procsignals;
    sigaction(SIGALRM, &actx, &oldactx);      
    }
  }
  
void doLocking (void)
  {
  if (ALLOW_SEVERAL_CONNS == 0) //NO simultaneous connections
    {
    TBuffer lockfile, temp;
    int res;
    strncpy (lockfile, ReturnUserLockFileName (username, temp), CMAXBUFFER);
    res = IsLocked (lockfile, &LockFileDescriptor);
    switch (res)
      {
      case 2: //NO EXISTE EL FICHERO DE LOCK
              LockFileDescriptor = CreateAndLockFile (lockfile);
              break;
      case 1: //SI HAY UN LOCK PA'FUERA
              LOG ("Lock file for user '%s'. Exiting.", username);
              html->ErrorPage (L->get(MSG_TOO_MAX_TRY_CONN)); 
              exit(0);        
              break;
      case 0: //NO HAY LOCK 
              int thepid;
              LOG ("User %s, IP=%s already connected. Killing old connection.", username, ipclient);  
              thepid = GetPIDFromLockFile(LockFileDescriptor);
              if (thepid >0) kill (thepid, SIGUSR1);
              EscribePIDenLockFile (LockFileDescriptor);
              break;
      }//switch
    }//if
  }   
  
void CheckCookie (UserOptions *uo, char *parm1, int cmd_counter, char *stlang, char *connid)
  {      
  if (Conf->getInt ("main", "use_cookies", 0) == 1)
    {
    //DEBUG ("parm1='%s',getValue='%s'", parm1, cookie->getValue());
    if (cmd_counter == 1)  //LA PRIMERA VEZ SI EL BROWSER NO SOPORTA/ACEPTA COOKIES ME LO DEVUELVE VACIO
      {
      if (strcmp (parm1, cookie->getValue()) != 0) uo->setWorkingWithCookies (false);
      else uo->setWorkingWithCookies (true);
      }
    else
      { 
      if (uo->IsWorkingWithCookies () == true) 
        {
        if (strcmp (parm1, cookie->getValue()) != 0) 
          {
          LOG ("SECURITY PROBLEM: Bad cookie from IP=%s, user=%s", ipclient, username);
          html->PrintBadCookiePage (stlang, username, connid, uo);
          }
        }
      }
    }
  }

void ClearCookie (UserOptions *uo)
  {
  if (Conf->getInt ("main", "use_cookies", 0) == 1)
    {
    if (uo->IsWorkingWithCookies () == true) 
      {
      html->printCookie (cookie->getSetClearFullCookieLine());
      }
    }
  }

//*******************************************************
//******************************************        LOGIN
bool doLogin (UserOptions *uo, char *imapserver, char *pw)
  {
  bool VALIDLOGIN = false;
  
  //ENTRY EMPTY IN CGI imapserver FIELD. WE CORRECT IT.
  if ((imapserver == NULL) || (strcmp (imapserver, "") == 0) || (strcmp (imapserver, " ") == 0))
    {
    strcpy (imapserver, "#");
    LOG ("Field IMAP server EMPTY. Forcing '#'. IP=%s, User '%s'.", ipclient, username);
    }
  //ENTRY # IN CGI imapserver FIELD AND WE MUST USE DB.
  if ((strcmp (imapserver, "#") == 0) && (Conf->getInt ("main", "usedbforconf", 0) == 1))
    {
    //USER DO NOT EXIST IN DB. DEFAULT TO SECTION def_user_not_in_db
    if (exist_user (Conf->getString ("main", "dbfile", "/var/postman/postman.db", CMAXBUFFER), username) == 0)
      {
      TBuffer xbuf;
      strncpy (xbuf, Conf->getString ("main", "def_user_not_in_db", "", CMAXBUFFER), CMAXBUFFER);
      AddSuffixToIPName (xbuf);
      uo->setImapServer (Conf->getString (xbuf, "imapserver", "imap.xx.xx", CMAXIPNAME));
      uo->setImapPort   (Conf->getInt    (xbuf, "imapport", DEFIMAPPORT));
      uo->setMailDomain (Conf->getString (xbuf, "maildomain", "xx.xx", CMAXIPNAME));
      uo->setSMTPServer (Conf->getString (xbuf, "smtpserver", "smtp.xx.xx", CMAXIPNAME));  
      uo->setNNTPServer (Conf->getString (xbuf, "nntpserver", "news", CMAXIPNAME));        
      cclient->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      uo->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      cclient->setPathToRemoteFolder (Conf->getString (xbuf, "remotepath", "", CMAXBUFFER));
      }
    //USER EXISTS IN DB
    else
      {
      TBuffer xbuf;
      int aimapport;
      IPNAME asmtpserver, amaildomain;
      db_get_data_from_user (Conf->getString ("main", "dbfile", "/var/postman/postman.db", CMAXBUFFER), username, 
                             imapserver, &aimapport, asmtpserver, amaildomain);
      strncpy (xbuf, imapserver, CMAXBUFFER);
      AddSuffixToIPName (xbuf);
      uo->setImapServer (xbuf);
      uo->setImapPort   (aimapport);
      uo->setSMTPServer (asmtpserver);
      uo->setMailDomain (amaildomain);
      uo->setNNTPServer (Conf->getString (xbuf, "nntpserver", "news", CMAXIPNAME));        
      cclient->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      uo->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      cclient->setPathToRemoteFolder (Conf->getString (xbuf, "remotepath", "", CMAXBUFFER));
      }        
    }    
  //ENTRY # IN CGI imapserver FIELD AND WE DO NOT MUST USE DB. DEFAULT TO SECTION def_user_not_in_db
  else if ((strcmp (imapserver, "#") == 0) && (Conf->getInt ("main", "usedbforconf", 0) == 0))
    {
    TBuffer xbuf;
    strncpy (xbuf, Conf->getString ("main", "def_user_not_in_db", "", CMAXBUFFER), CMAXBUFFER);
    AddSuffixToIPName (xbuf);
    uo->setImapServer (Conf->getString (xbuf, "imapserver", "imap.xx.xx", CMAXIPNAME));
    uo->setImapPort   (Conf->getInt    (xbuf, "imapport", DEFIMAPPORT));
    uo->setMailDomain (Conf->getString (xbuf, "maildomain", "xx.xx", CMAXIPNAME));
    uo->setSMTPServer (Conf->getString (xbuf, "smtpserver", "smtp.xx.xx", CMAXIPNAME));  
    uo->setNNTPServer (Conf->getString (xbuf, "nntpserver", "news", CMAXIPNAME));        
    cclient->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
    uo->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
    cclient->setPathToRemoteFolder (Conf->getString (xbuf, "remotepath", "", CMAXBUFFER));
    }        
  //TEXT ENTRY IN CGI imapserver FIELD
  else
    {
    TBuffer xbuf;
    strncpy (xbuf, imapserver, CMAXBUFFER);
    AddSuffixToIPName (xbuf);
    //EXISTS imapserver SECTION
    if (Conf->IsThereSection (xbuf) == true)
      {
      uo->setImapServer (Conf->getString (xbuf, "imapserver", "imap.xx.xx", CMAXIPNAME));
      uo->setImapPort   (Conf->getInt    (xbuf, "imapport", DEFIMAPPORT));
      uo->setMailDomain (Conf->getString (xbuf, "maildomain", "xx.xx", CMAXIPNAME));
      uo->setSMTPServer (Conf->getString (xbuf, "smtpserver", uo->getImapServer(), CMAXIPNAME));  
      uo->setNNTPServer (Conf->getString (xbuf, "nntpserver", "news", CMAXIPNAME));        
      cclient->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      uo->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
      cclient->setPathToRemoteFolder (Conf->getString (xbuf, "remotepath", "", CMAXBUFFER));
      }
    //DO NOT EXIST imapserver SECTION IN interdaemon.cfg FILE
    else
      {
      //WE DO NOT ALLOW LOGINS IF imapserver SECTION DO NOT EXIST
      if (Conf->getInt ("main", "allowotherimapservers", 0) == 0)
        {
        LOG ("IMAP server '%s' is not allowed. IP=%s, User=%s. Exiting.", imapserver, ipclient, username);
        html->ErrorPage (L->get(L_NOTALLOWED_IMAPSERVER));
        exit(0);
        }
      //WE ALLOW LOGINS TO EVERY imapserver, BUT OVERWRITING SOME FIELDS
      else
        {
        uo->setImapServer (xbuf);
        if (strcmp (Conf->getString ("*", "imapport", "*", CMAXIPNAME), "*") == 0)  
          {uo->setImapPort (DEFIMAPPORT);}
        else 
          {uo->setImapPort (Conf->getInt ("*", "imapport", DEFIMAPPORT));}
        if (strcmp (Conf->getString ("*", "smtpserver", "*", CMAXIPNAME), "*") == 0)  
          {uo->setSMTPServer (imapserver);}
        else 
          {uo->setSMTPServer (Conf->getString ("*", "smtpserver", "smtp.xx.xx", CMAXIPNAME));}                                  
        if (strcmp (Conf->getString ("*", "nntpserver", "*", CMAXIPNAME), "*") == 0)  
          {uo->setNNTPServer (imapserver);}
        else 
          {uo->setNNTPServer (Conf->getString ("*", "nntpserver", "news", CMAXIPNAME));}                                  
        if (strcmp (Conf->getString ("*", "maildomain", "*", CMAXIPNAME), "*") == 0)  
          {uo->setMailDomain (imapserver);}
        else                       
          {uo->setMailDomain (Conf->getString ("*", "maildomain", "xx.xx", CMAXIPNAME));}                                                  
        cclient->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
        uo->setMailboxPrefix (Conf->getString (xbuf, "mailboxprefix", "", CMAXBUFFER));
        cclient->setPathToRemoteFolder (Conf->getString (xbuf, "remotepath", "", CMAXBUFFER));
        }
      }
    }   
  sm_setproctitle (PROGNAME, true, "%s   (Before IMAPValidate)", username);  
  VALIDLOGIN = cclient->IMAPValidate (uo->getImapServer(), uo->getImapPort(), "INBOX", username, pw);
  sm_setproctitle (PROGNAME, true, "%s   (After IMAPValidate)", username);  
  cclient->setNNTPServer (uo->getNNTPServer());
  return VALIDLOGIN;
  }

//SWITCH ON THE ALARM
void SwitchOnAlarm (UserOptions *uo, int alarmtime)
  {
  //DEBUG ("--%d--", alarmtime);
  uo->setTimeout (alarmtime);
  if (USETIMEOUT == 1) alarm ((xuint)(alarmtime));
  }

void ExitServer (UserOptions *uo)
  {
  serversocket->Close ();
  delete serversocket;
  delete uo;  
  delete cclient;
  delete L;
  delete html;
  if (getpid () == MainServerPid) {LOG ("Main server is down.");}
  CLOSELOG();
  exit (0);
  }

//ERROR: PW invalida, server not ready, port not open, ...
void ErrorInLogin (void)
  {
  TBuffer errmsg;
  strncpy (errmsg, cclient->getErrorString(), CMAXBUFFER);
  LOG ("%s (ipclient='%s', user='%s')", errmsg, ipclient, username);
  html->ErrorPage(errmsg);
  }

void setUserDirs (UserOptions *uo)
  {
  TBuffer asb, temp;
  char inicials[3];    
  strncpy (UserDir, ReturnUsersDir(Conf->getString ("main", "usersbasepath", "/var/postman/users", CMAXBUFFER), username, uo->getImapServer (), temp), CMAXBUFFER);
  inicials [0] = username[0]; inicials [1] = username[1]; inicials [2] = '\0';
  strncpy (asb, Conf->getString ("main", "usersbasepath", "/var/postman/users", CMAXBUFFER), CMAXBUFFER);strcat (asb, "/"); strcat (asb, uo->getImapServer ());
  if (CreateDir (asb, USERDIRMODE) == false) {html->ErrorPage(L->get (ERR_CREAT_USERDIR)); fflush (stdout); exit (0);}
  strcat (asb, "/"); strcat (asb, inicials);
  if (CreateDir (asb, USERDIRMODE) == false) {html->ErrorPage(L->get (ERR_CREAT_USERDIR)); fflush (stdout); exit (0);}
  if (CreateDir (UserDir, USERDIRMODE) == false) {html->ErrorPage(L->get (ERR_CREAT_USERDIR)); fflush (stdout); exit (0);}    
  if (CreateDir ((char *)uo->getAttachDir(), USERATTACHSDIRMODE) == false) {html->ErrorPage(L->get (ERR_CREAT_USERATTACHSDIR)); fflush (stdout); exit (0);}   
  }

//WE ADD A DEFAULT SUFFIX TO IMAPSERVER IF THIS HAVE NOT IT (post -> post.uv.es).
void AddSuffixToIPName (char *server)
  {
  if (strchr (server, '.') == NULL) 
    {
    strcat (server, Conf->getString ("main", "defaultipdomain", "", CMAXIPNAME)); 
    }
  }   

