/*
UserOptions.h
*/

#ifndef __USEROPTIONS_H
#define __USEROPTIONS_H

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <signal.h>

#include "Utils.h"
#include "Config.h"
#include "Configuration.h"
#include "XString.h"
#include "XVector.h"
#include "CCLIENT.h"

typedef XVector<XString> StringList;

#define DEF_USEROPT_FILE "\n[main]\n"                                                        \
                         "msgsperindexpage="STR(DEFOPT_MSGSPERINDEXPAGE)"\n"                 \
                         "personalname=\n"                                                   \
                         "insigninnewmsg="STR(DEFOPT_INSERTSIGNATUREINNEWMSG)"\n"            \
                         "savemsgsentmail="STR(DEFOPT_SAVEMSGSENTMAIL)"\n"                   \
                         "replyfrom="STR(DEFOPT_REPLYFROM)"\n"                               \
                         "forwattachs="STR(DEFOPT_FORW_ATTACHS)"\n"                          \
                         "heightwritearea="STR(DEFOPT_HEIGHTWRITEAREA)"\n"                   \
                         "widthwritearea="STR(DEFOPT_WIDTHWRITEAREA)"\n"                     \
                         "truncatelengthreadingmsg="STR(DEFOPT_TRUNCATELENGTHREADINGMSG)"\n" \
                         "confirmpurge="STR(DEFOPT_CONFIRMPURGE)"\n"                         \
                         "confirmlogout="STR(DEFOPT_CONFIRMLOGOUT)"\n"                       \
                         "skin="DEFOPT_SKIN"\n"                                              \
                         "refreshtime="STR(DEFOPT_REFRESHTIME)"\n"                           \
                         "sentmailname="DEFOPT_SENTMAILBOX"\n\n"

class UserOptions
  {
  private:
    Configuration *TheConf;
    TBuffer StatusLine, LastCmd;
    XString user, imapserver, basepath, userdir, userattdir, fnuseroptions, fnsignature, 
      fnsubscribedgroups, fnrcgroups, fnsavedmsg, fnsavednntpmsg, SentmailName, 
      XSTrueFileNameAttach;
    bool InsertSignatureInNewMsg, SaveMsgSentMail, ForwAttachs, ReplyFrom, WorkingWithCookies, UseJavaScript,
         ConfirmPurge, ConfirmLogout;
    long MsgsPerIndexPage, CurrentPosition, TotalMsgs, MAILBOXSIZE, BeginPage;
    XString personalname, signature;
    int msgtoanswer, validator, pageid, lastcmd, sort_ab, TruncateLengthReadingMsg, WidthWriteArea, 
        HeightWriteArea, RefreshTime, IMAPPORT, timeout;
    StringList SLAttachList, SLAttachListInfo, PilaCmds;
    IPNAME MAILDOMAIN, IMAPSERVER, SMTPSERVER, NNTPSERVER;
    TBuffer MAILBOXPREFIX;
    char skin[MAXLENGTHSKIN + 1];
    COLORHTML COLOR_BGTABLE, COLOR_BGTABLECOPYRIGHT, COLOR_BGBOTONERA, COLOR_BGTABLEMSGINDEX, COLOR_BGBODY,
              COLOR_BGTEXTO, COLOR_BGFILASMSGINDEX, COLOR_BGFILASADDBOOK, COLOR_BGFILASREADINGMSG,
              COLOR_FONT_TABLE, COLOR_FONT_MAIN, COLOR_FONT_STATUS, COLOR_FONT_STATUSLINE;
    FONTNAME FONT_TABLE, FONT_MAIN, FONT_STATUS;
    FONTSIZE FONT_TABLE_SIZE, FONT_MAIN_SIZE, FONT_STATUS_SIZE, FONT_BOTONERA_SIZE;
    TSBuffer IMG_BG;
  public:
    StringList *SLSubscribedGroups;
    UserOptions (const char *abasepath, const char *auser, const char *aimapserver);
    virtual ~UserOptions (); 
    void leeConf ();
    void InitStatusLine (void);
    const char *getStatusLine (void);
    void setStatusLine (const char *newstatusline, const char *htmlcolor);
    int getValidator (void);
    void IncValidator (void);
    void setCurrentPosition (long pos, long total);
    long getCurrentPosition (void);
    long getBeginPage (void);    
    void setMailboxSize (long asize);
    long getMailboxSize (void);    
    long getTotalMsgs (void);
    void setInsertSignatureInNewMsg (bool value);
    bool getInsertSignatureInNewMsg (void); 
    void setMailDomain (const char *amaildomain);
    const char *getMailDomain (void);
    void setImapServer (const char *aimapserver);
    const char *getImapServer (void);                  
    void setImapPort (int aimapport);
    int getImapPort (void);                  
    void setSMTPServer (const char *asmtpserver);
    const char *getSMTPServer (void);  
    void setNNTPServer (const char *anntpserver);      
    const char *getNNTPServer (void);
    void setMailboxPrefix (const char *amailboxprefix);
    const char *getMailboxPrefix (void);
    const char *getUsername (void);
    const char *getAttachDir (void);    
    void setMsgsPerIndexPage (long value);
    long getMsgsPerIndexPage (void);
    void setPersonalName (char *value);
    const char *getPersonalName (void);
    bool getSaveMsgSentMail (void);
    void setSaveMsgSentMail (bool value);
    void setReplyFrom (bool value);
    bool getReplyFrom (void);
    void setForwAttachs (bool value);
    bool getForwAttachs (void);
    void setTruncateLengthReadingMsg (int value);
    int getTruncateLengthReadingMsg (void);
    void setWidthWriteArea (int value);
    int  getWidthWriteArea (void);    
    void setHeightWriteArea (int value);
    int  getHeightWriteArea (void);            
    void setRefreshTime (int value);
    int  getRefreshTime (void);            
    const char *getSignature (void);
    bool SaveOptions (char *tmpfilename);
    bool SaveNNTPOptions (char *tmpfilename);
    void LoadOptions (void);
    int numAttachs (void);
    const char *getFileNameAttach (int p);
    const char *getTrueFileNameAttach (int p);
    const char *getAttachInfo (int p);
    int addFileNameAtach (const char *newfile, const char *info);
    int delFileNameAtach (const char *attachnumbers);
    void delAllAttachs (void);
    void setSort_ab (int value);
    int getSort_ab (void);    
    const char *getfnSavedMsg (void);
    void setLastCmd (int value);
    int getLastCmd (void);        
    void vuelcaPila (void);
    void setPageId (int value);
    int getPageId (void);    
    const char *getLastStringCmd (int cmdactual, const char *lang, const char *connid, bool withhtml);    
    void addLastStringCmd (const char *stcmd);
    void setNumMsgToReply (int value);
    int getNumMsgToReply(void);    
    void setSkin (const char *askin);
    const char *getSkin (void);
    char *getSkinAtribProp (const char *propiedad);
    int getSkinAtribPropI (const char *propiedad);
    bool IsWorkingWithCookies (void);
    void setWorkingWithCookies (bool b);
    bool get_use_javascript (void);
    void set_use_javascript (bool b);
    const char *getSentMailName (void);
    void setSentMailName (const char *smname);
    bool getConfirmPurge (void);
    void setConfirmPurge (bool b);
    bool getConfirmLogout (void);
    void setConfirmLogout (bool b);    
    void setTimeout (int value);
    int getTimeout (void);        
    void initSubscribedGroups();
    const char *getFileNameRcGroups (void);    
    int getObligatedGroups(TBuffer obligatedgroups);    
    bool SaveSubscribedGroups(void);
  };

#endif

